/**
 * $Id: connqueue.c 1 2006-04-29 18:51:20Z alexeen $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <netinet/in.h>
#include <assert.h>

#include "connqueue.h"
#include "common.h"

/**
 * cq_init: Initializes a connection queue.
 * @param cq - The connection queue.
 */
void cq_init(connqueue_t *cq)
{
	assert(cq);

	cq->head = NULL;
	cq->tail = NULL;

  /** Initializes mutex. **/
	cq->cq_mta = (pthread_mutexattr_t *) malloc(sizeof(pthread_mutexattr_t));
	cq->cq_mutex = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
	pthread_mutexattr_init(cq->cq_mta);
	pthread_mutex_init(cq->cq_mutex, cq->cq_mta);
}

/**
 * cq_enqueue: Enqueue a socket descriptor
 *   into our connection queue.
 * @param sock - The socket descriptor to enqueue.
 * @param addr - The socket address.
 * @param cq - The connection queue.
 */
void cq_enqueue(int sock, struct sockaddr_in *addr, connqueue_t *cq)
{
	connqueue_elm_t *elm = NULL;

	assert(cq);
	assert(addr);
	assert(0 <= sock);
	assert(!(NULL == cq->head && NULL != cq->tail));
	assert(!(NULL != cq->head && NULL == cq->tail));

	/** Allocate memory for our element. **/
	elm = (connqueue_elm_t *)calloc(sizeof(connqueue_elm_t), 1);
	elm->sock = sock;
	elm->addr = addr;
	elm->next = NULL;

	if (NULL == cq->head && NULL == cq->tail)
	{
		/** No existing elements in queue. **/
		cq->head = elm;
		cq->tail = elm;
	}
	else
	{
		/** Append element to tail of queue. **/
		cq->tail->next = elm;
		cq->tail = elm;
	}
}

/**
 * cq_dequeue: Dequeues a socket descriptor
 *   from the connection queue.
 * @param sockaddr - Socket address storage.
 * @param cq - The connection queue.
 * @return Returns a socket descriptor. Returns -1
 *   if queue is empty.
 */
int cq_dequeue(struct sockaddr_in *addr, connqueue_t *cq)
{
	int retval = -1;
	connqueue_elm_t *elm = NULL;

	assert(cq);
	assert(addr);
	assert(!(NULL == cq->head && NULL != cq->tail));
	assert(!(NULL != cq->head && NULL == cq->tail));

	if (NULL == cq->head && NULL == cq->tail)
	{
		/** Queue is empty. **/
		return -1;
	}

	/** Dequeue from the head. **/
	elm = cq->head;
	assert(0 <= elm->sock);
	assert(elm->addr);

	retval = elm->sock; /** Grab FD. **/

	/** Make a copy of the socket address. **/
	memcpy(addr, elm->addr, sizeof(struct sockaddr_in));

	if (cq->head == cq->tail)
	{
		/** Last element dequeued. **/
		cq->head = NULL;
		cq->tail = NULL;
	}
	else
	{
		cq->head = elm->next; /** Move head to next in queue. **/
	}

	/** Free resources allocated for element. **/
	free(elm->addr); /** No longer need this. **/
	free(elm);
	return retval;
}

/**
 * cq_is_empty: Checks if the queue is empty.
 * @param cq - The connection queue.
 * @return Returns 1 if it is empty, 0 otherwise.
 */
int cq_is_empty(connqueue_t *cq)
{
	assert(cq);
	assert(!(NULL == cq->head && NULL != cq->tail));
	assert(!(NULL != cq->head && NULL == cq->tail));

	if (NULL == cq->head && NULL == cq->tail)
	{
		return 1;
	}

	return 0;
}

/**
 * cq_destroy: Free allocated resources for connection
 *  queue data structure.
 * @param cq - The connection queue.
 */
void cq_destroy(connqueue_t *cq)
{
	assert(cq);

  /** Destroy mutex. **/
	pthread_mutexattr_destroy(cq->cq_mta);
	pthread_mutex_destroy(cq->cq_mutex);

  /** Free allocated resources. **/
  free(cq->cq_mta);
  free(cq->cq_mutex);
  free(cq);

	cq->head = NULL;
	cq->tail = NULL;
}
