/**
 * $Id: connrelay.c 11 2006-05-01 23:58:23Z alexeen $
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>
#include <assert.h>

#include "workers.h"
#include "connqueue.h"
#include "connrelay.h"
#include "sendall.h"
#include "opts.h"
#include "common.h"

#ifndef NIDENT /** Ident support. **/
#include <ident.h>
#endif

extern char listening;    /** Keeps track if program should continue relaying. **/
extern workforce_t *g_wf; /** Workforce worker thread pool. **/
extern connqueue_t *g_cq; /** Connection queue. **/
extern opts_t *g_opts;    /** Runtime options. **/
extern struct hostent *rhostent; /** Resolved remote host network address. **/

static void _do_relay(int csock, struct sockaddr_in *caddr);

/**
 * connrelay: Connection relay handler.
 */
void *connrelay()
{
	int csock = -1;
	struct sockaddr_in caddr;

	assert(g_wf);
	assert(g_cq);
	assert(g_opts);
	assert(rhostent);

	while (1)
	{
		pthread_mutex_lock(g_wf->wf_mutex);
		pthread_mutex_lock(g_cq->cq_mutex);
		while (!WF_IS_TERM(g_wf->terminate) && 0 > (csock = cq_dequeue(&caddr, g_cq)))
		{
			/** Keep waiting for pending tasks. **/
			pthread_mutex_unlock(g_cq->cq_mutex);
			pthread_cond_wait(g_wf->wf_cv, g_wf->wf_mutex);
			pthread_mutex_lock(g_cq->cq_mutex);
		}

		/** Terminate immediately if required. **/
		if (WF_IS_TERM(g_wf->terminate))
		{
			pthread_mutex_unlock(g_cq->cq_mutex);
			pthread_mutex_unlock(g_wf->wf_mutex);
			break;
		}

		assert(0 <= csock);

		pthread_mutex_unlock(g_cq->cq_mutex);
		pthread_mutex_unlock(g_wf->wf_mutex);

		_do_relay(csock, &caddr);
	}

	/** Explicitly terminate the thread. **/
	pthread_exit(NULL);	
}

/**
 * _do_relay: Handles relaying data between client and
 *  remote server.
 * @param csock - The client socket to be handled.
 * @param caddr - Client's socket address.
 */
static void _do_relay(int csock, struct sockaddr_in *caddr)
{
	int rsock = -1;   /** Remote server socket descriptor. **/
	int fdmax = -1;   /** Largest socket descriptor value among sets. **/
	fd_set master;    /** Master list of file descriptors. **/
	fd_set read_fds;  /** List of readable file descriptors. **/
	struct sockaddr_in r_addr;  /** Remote host's Internet address. **/
	char buf[BUFLEN]; /** Buffer data passing through relay. **/
	int nbytes;       /** Number of bytes read/write. **/
	int retval;				/** Return value storage. **/
	struct timeval tv; /** Select timeout value. **/

#ifndef NIDENT
	char *idnt_reply = NULL; /** Store the ident reply. **/
#endif

	assert(0 <= csock);
	assert(caddr);

	if (-1 == (rsock = socket(PF_INET, SOCK_STREAM, 0)))
	{
#ifndef NDEBUG
perror("Failed to acquire socket for connecting to remote server!");
#endif
		goto EDONE;
	}

	r_addr.sin_family = rhostent->h_addrtype; /** Host byte order (AF_INET). **/
	r_addr.sin_port = htons(g_opts->rport); /** Remote port in short network byte order. **/
	/** Copy remote host address. **/
	memcpy(&r_addr.sin_addr, rhostent->h_addr, sizeof(struct in_addr));
	memset(&(r_addr.sin_zero), '\0', 8); /** Zero the rest of the struct. **/

	if (-1 == connect(rsock, (struct sockaddr *)&r_addr, sizeof(struct sockaddr)))
	{
#ifndef NDEBUG
perror("Failed to connect to remote host!");
#endif
		goto EDONE;
	}

	/** Clear file descriptor sets. **/
	FD_ZERO(&master);
	FD_ZERO(&read_fds);

	/** Place client socket in master set. **/
	FD_SET(csock, &master);
	FD_SET(rsock, &master);

	/** Largest FD value. **/
	fdmax = csock < rsock ? rsock : csock;

#ifndef NIDENT
	/** If ident is required. **/
	if (IS_IDENT(g_opts->flags))
	{
		/** Query for IDENT reply. **/
		idnt_reply = ident_id(csock, IDENT_TIMEOUT);
		if (!idnt_reply)
		{
			/** Fail to get IDENT. **/
			idnt_reply = NULL;
			nbytes = 27;
			memset(buf, '\0', BUFLEN); /** Clear data buffer. **/
			strncpy(buf, "530 Failed to get IDENT!\r\n", BUFLEN);
			sendall(csock, buf, nbytes, &retval);
			goto DONE;
		}
		else
		{
			memset(buf, '\0', BUFLEN); /** Clear data buffer. **/
			snprintf(buf, BUFLEN, "IDNT %s@%s:%s\n", idnt_reply,
				inet_ntoa(caddr->sin_addr),
				inet_ntoa(caddr->sin_addr));

			buf[BUFLEN] = '\0'; /** Avoid buffer overflow. **/

			if (-1 == sendall(rsock, buf, strlen(buf), &retval))
			{
				/** Failed to send IDENT to server. **/
				free(idnt_reply);
				idnt_reply = NULL;
				nbytes = 38;
				memset(buf, '\0', BUFLEN); /** Clear data buffer. **/
				strncpy(buf, "421 Unable to establish connection!\r\n", BUFLEN);
				sendall(csock, buf, nbytes, &retval); /** Don't care if client receives it. **/
				goto DONE;
			}
		}

		/** Don't need the reply anymore. **/
		free(idnt_reply);
		idnt_reply = NULL;
	}
#endif /** NIDENT **/

	memset(buf, '\0', BUFLEN); /** Clear data buffer. **/

	while (001 == (001 & listening))
	{
		/** Select timeout value resets. **/
		tv.tv_sec = SEL_TO_S;
		tv.tv_usec = SEL_TO_US;

		read_fds = master; /** Copy from master set every time. **/

		if (-1 == select(fdmax+1, &read_fds, NULL, NULL, &tv))
		{
#ifndef NDEBUG
perror("Select system call failed!");
#endif
			break; /** Stop relaying immediately. **/
		}

		/** Check both descriptors to see if there's any into to relay. **/
		if (FD_ISSET(csock, &read_fds))
		{
			/** Client wants to send server data. **/
			if (-1 == (nbytes = recv(csock, buf, BUFLEN-1, 0)))
			{
#ifndef NDEBUG
perror("Failed to receive data from client socket!");
#endif
				break;
			}

			if (nbytes < 1) { break; } /** Nothing read. **/

			buf[nbytes] = '\0'; /** Prevent buffer overflow attacks. **/

#ifdef __SPY__ /** Spy on conversations in debug mode. **/
			if (IS_DEBUG(g_opts->flags))
			{
				fprintf(stderr, "C->S: %s", buf);
			}
#endif

			if (-1 == sendall(rsock, buf, nbytes, &retval))
			{
#ifndef NDEBUG
perror("Failed to send data to server socket!");
#endif
				break;
			}

			memset(buf, '\0', BUFLEN); /** Clear data buffer. **/
		}

		if (FD_ISSET(rsock, &read_fds))
		{
			/** Server wants to send client data. **/
			if (-1 == (nbytes = recv(rsock, buf, BUFLEN-1, 0)))
			{
#ifndef NDEBUG
perror("Failed to receive data from server socket!");
#endif
				break;
			}

			if (nbytes < 1) { break; } /** Nothing read. **/

			buf[nbytes] = '\0'; /** Prevent buffer overflow attacks. **/

#ifdef __SPY__ /** Spy on conversations in debug mode. **/
			if (IS_DEBUG(g_opts->flags))
			{
				fprintf(stderr, "S->C: %s", buf);
			}
#endif

			if (-1 == sendall(csock, buf, nbytes, &retval))
			{
#ifndef NDEBUG
perror("Failed to send data to client socket!");
#endif
				break;
			}

			memset(buf, '\0', BUFLEN); /** Clear data buffer. **/
		}
	}

	/** Perform clean up. **/
#ifndef NIDENT
DONE:
	if (NULL != idnt_reply) { free(idnt_reply); }
#endif
	FD_CLR(csock, &master);
	FD_CLR(rsock, &master);
	close(rsock);

EDONE:
	close(csock);
}
