/**
 * $Id: opts.c 1 2006-04-29 18:51:20Z alexeen $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <assert.h>

#include "opts.h"
#include "common.h"

/**
 * opts_init: This method simply initializes the opts_t
 *   structure in memory.
 * @param opts - The opts_t structure storage memory.
 */
void opts_init(opts_t *opts)
{
	assert(opts);

	/** Make it clean. **/
	memset(opts, '\0', sizeof(opts_t));

	/** Assume normal execution by default. **/
	SET_NORMAL(opts->flags);

	opts->lport = -1;
	opts->rport = -1;
}

/**
 * opts_get: Parses command line arguments and
 *   set appropriate runtime flags in the opts_t
 *   data structure.
 * @param opts - The opts_t data struct storage memory.
 * @param argc - The number of arguments to parse.
 * @param argv - The arguments list to parse.
 * @return Returns -1 if errored, 0 for success.
 */
int opts_get(opts_t *opts, int argc, char **argv)
{
	int c = -1; /** Short argument value.  **/
	int option_index = 0; /** Stores current option index. **/

	assert(opts);
	assert(0 < argc);
	assert(argv);

	/** Process all arguments. **/
	while (1)
	{
		option_index = 0; /** Reset option index. **/

		/** Long arguments mapping to short arguments. **/
		static struct option long_options[] = {
			{"help", 0, 0, 'h'},
			{"debug", 0, 0, 'd'},
#ifndef NIDENT
			{"ident", 0, 0, 'i'},
#endif
			{"listen", 1, 0, 'l'},
			{"port", 1, 0, 'p'},
			{"server", 1, 0, 's'},
			{"version", 0, 0, 'v'},
			{0, 0, 0, 0}
		};

		/** Acquire option. **/
#ifndef NIDENT
		c = getopt_long(argc, argv, "hidl:p:s:v", long_options, &option_index);
#else
		c = getopt_long(argc, argv, "hdl:p:s:v", long_options, &option_index);
#endif

		/** Finish processing all arguments. **/
		if (-1 == c) { break; }

		/** Process current argument. **/
		switch (c)
		{
			case 'h': /** Display usage information. **/
				SET_HELP(opts->flags);
				break;

#ifndef NIDENT
			case 'i': /** Pass IDENT to server, mainly for FTP. **/
				SET_IDENT(opts->flags);
				break;
#endif

			case 'd': /** Enable debug mode. **/
				SET_DEBUG(opts->flags);
				break;

			case 'l': /** Port to listen on. **/
				opts->lport = (int) strtol(optarg, (char **)NULL, 10);
				break;

			case 'p': /** Remote port to connect to. **/
				opts->rport = (int) strtol(optarg, (char **)NULL, 10);
				break;

			case 's':	/** Remote server to connect to. **/
				strncpy(opts->rhost, optarg, BUFLEN-1);
				opts->rhost[BUFLEN-1] = '\0'; /** Avoid buffer overflow. **/
				break;

			case 'v': /** Display version information. **/
				SET_VERSION(opts->flags);
				break;

			case '?': /** Unknown option. **/
				break;

			default: /** getopt_long returned error character code. **/
				fprintf(stderr, "?? getopt returned character code 0%o ??\n", c);
				return -1;
		}
	}

	/** Check required arguments. **/
	if (1 > opts->lport || 65534 < opts->lport ||
		1 > opts->rport || 65534 < opts->rport ||
		1 > strlen(opts->rhost))
	{
		SET_HELP(opts->flags);
	}

	return 0;
}
