/**
 * $Id: tcprelay.c 10 2006-05-01 03:55:01Z alexeen $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>
#include <assert.h>

#include "resolvehost.h"
#include "opts.h"
#include "workers.h"
#include "connqueue.h"
#include "connrelay.h"
#include "droproot.h"
#include "sighandler.h"
#include "common.h"

/** Global variables. **/
opts_t *g_opts = NULL;           /** Runtime options. **/
workforce_t *g_wf = NULL;        /** Workforce workers thread pool. **/
connqueue_t *g_cq = NULL;        /** Connection queue. **/
struct hostent *rhostent = NULL; /** Remote host network names and addresses. **/
char *rhostent_buf = NULL;       /** Used to buffer rhostent results. **/

/** Keeps track of when to stop listening. **/
char listening = 001;

static void version(char *prog);
static void usage(char *prog);
static int setup_globals();
static void destroy_globals();
static int do_listen();
static pid_t daemonize();

#if defined(HIDE_PROC_TITLE) && defined(FAKE_PROC_NAME) && defined(FAKE_PROC_TITLE)
#include "setproctitle.h"
static void set_proc_title(int argc, char **argv);
#endif

/**
 * Where life of our process begins. ;)
 */
int main(int argc, char **argv)
{
	int herr = -1; /** Store resolution return status. **/
	pid_t pid;     /** PID returned by fork. **/

	/** Allocate memory and initialize runtime options. **/
	if (NULL == (g_opts = (opts_t *)malloc(sizeof(opts_t))))
	{
#ifndef NDEBUG
perror("Out of memory");
#endif
		return EXIT_FAILURE;
	}
	opts_init(g_opts);

	/** Parse and store options. **/
	if (opts_get(g_opts, argc, argv))
	{
#ifndef NDEBUG
perror("Failed to parse and set runtime options");
#endif
		return EXIT_FAILURE;
	}

	/** Take care of version and usage display. **/
	if (IS_VERSION(g_opts->flags)) { version(argv[0]); goto EXIT; }
	else if (IS_HELP(g_opts->flags)) { usage(argv[0]); goto EXIT; }

	if (IS_DEBUG(g_opts->flags))
	{
		printf("Listening on port: %d\n", g_opts->lport);
		printf("Connect to port: %d\n", g_opts->rport);
		printf("Connect to host: %s\n", g_opts->rhost);
#ifndef NIDENT
		printf("Pass Ident: %s\n", (IS_IDENT(g_opts->flags) ? "Enabled" : "Disabled"));
#endif
		printf("Run Time Option Flag: 0x%X\n", g_opts->flags);
	}

/** Does not fork into background if compiled to display debug messages. **/
#ifdef NDEBUG
	/** Fork into daemon process. **/
	if (!IS_DEBUG(g_opts->flags))
	{
#if defined(HIDE_PROC_TITLE) && defined(FAKE_PROC_NAME) && defined(FAKE_PROC_TITLE)
		set_proc_title(argc, argv); /** Change the process title. **/
#endif
		pid = daemonize();
		if (pid < 0) { goto EXIT; }      /** Failed to fork. **/
		else if (pid > 0) { goto EXIT; } /** Parent can exit after successful fork. **/
		/** Child will continue execution. **/
	}
	else
	{
		/** Setup signal handling for this foreground process. **/
		setup_sighandlers();
	}
#endif

	/** Resolve hostname/IP address entered. **/
	rhostent_buf = (char *)malloc(BUFLEN);
	if (NULL == rhostent_buf)
	{
#ifndef NDEBUG
perror("Out of memory");
#endif
		goto EXIT;
	}

	rhostent = (struct hostent *)malloc(sizeof(struct hostent));
	if (NULL == rhostent)
	{
#ifndef NDEBUG
perror("Out of memory");
#endif
		goto EXIT;
	}

	herr = resolve_host(g_opts->rhost, rhostent, rhostent_buf, BUFLEN);
	if (herr < 0)
	{
		/** Failed to resolve host address. **/
		if (IS_DEBUG(g_opts->flags))
		{
			printf("Failed to resolve remote host address: %s\n", g_opts->rhost);
		}
		goto ERROR;
	}

	if (setup_globals()) { goto ERROR; }

	/** Start listening and accepting connections. **/
	if (do_listen()) { goto ERROR; }

EXIT:
	destroy_globals();
	if (NULL != rhostent_buf) { free(rhostent_buf); }
	if (NULL != rhostent) { free(rhostent); }
	free(g_opts);
	return EXIT_SUCCESS;

ERROR:
	destroy_globals();
	if (NULL != rhostent_buf) { free(rhostent_buf); }
	if (NULL != rhostent) { free(rhostent); }
	free(g_opts);
	return EXIT_FAILURE;
}

/**
 * version: Display program version.
 * @param prog - Program name.
 */
static void version(char *prog)
{
	assert(prog);
	printf("%s, version %s\n", prog, VERSION);
}

/**
 * usage: Display usage information.
 * @param prog - Program name.
 */
static void usage(char *prog)
{
	assert(prog);
	printf("%s, version %s\n\n", prog, VERSION);
	printf("usage: %s [-dhv] -l <listen port> -p <remote port> -s <remote host>\n", prog);
	printf("  d, --debug   : debug mode (execution in foreground)\n");
#ifndef NIDENT
	printf("  i, --ident   : enable pass IDENT to remote server\n");
#endif
	printf("  h, --help    : display usage information (this page)\n");
	printf("  l, --listen  : port to listen on for accepting connections\n");
	printf("  p, --port    : remote port to connect to\n");
	printf("  s, --server  : remote server to connect to\n");
	printf("  v, --version : display version information\n\n");
	printf("Author: Lee, W.C. Alex\n");
	printf("Report Bugs: <alexeen@noservice.net>\n");
}

/**
 * daemonize: Fork process into daemon.
 * @return Returns the pid of the current process.
 */
static pid_t daemonize()
{
	pid_t pid, sid;

	pid = fork(); /** Fork a new process. **/

	if (0 > pid)
	{
#ifndef NDEBUG
perror("Failed to fork into daemon");
#endif
		return pid;
	}

	if (0 < pid)
	{
		/** Parent can exit now. **/
		return pid;
	}

	/** Change the file mode mask. **/
	umask(0);

	/** Create a new SID for the child process. **/
	sid = setsid();

	if (0 > sid)
	{
#ifndef NDEBUG
perror("Forked child fail to acquire new session");
#endif
		return sid;
	}

	/** Drop all root privileges if running as privileged user. **/
	if (drop_root()) { return -1; }

	/** Change to a path that exists for sure. **/
	if (0 > (chdir("/")))
	{
		/** Root directory does not exist, THIS IS VERY BAD! **/
		return -1;
	}

	/** Closing all the standard file descriptors. **/
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);

	/** Setup signal handling for this daemon process. **/
	setup_sighandlers();

	return pid;
}

/**
 * setup_globals: Initializes global variables.
 * @return Returns 0 for success, -1 otherwise.
 */
static int setup_globals()
{
	/** Initialize workforce. **/
	if (NULL == (g_wf = (workforce_t *)calloc(sizeof(workforce_t), 1)))
	{
#ifndef NDEBUG
perror("Out of memory");
#endif
		return -1;
	}

	/** Initialize connection queue. **/
	if (NULL == (g_cq = (connqueue_t *)calloc(sizeof(connqueue_t), 1)))
	{
#ifndef NDEBUG
perror("Out of memory");
#endif
		free(g_wf); /** Manually free the memory. **/
		return -1;
	}

	cq_init(g_cq); /** Initializes connection queue. **/

	if (wf_init(MAXWORKERS, g_wf, &connrelay))
	{
#ifndef NDEBUG
perror("Failed to initialize workforce");
#endif
		destroy_globals();
		return -1;
	}

	return 0;
}

/**
 * destroy_globals: Destroy global variables and free allocated
 *   resources.
 */
static void destroy_globals()
{
	if (NULL != g_wf) { wf_destroy(g_wf); }
	if (NULL != g_cq) { cq_destroy(g_cq); }
}

/**
 * do_listen: Listens for incoming TCP connections.
 * @return Returns 0 if gracefully terminated, 1 if errored.
 */
static int do_listen()
{
	fd_set master;       /** Master list of file descriptors. **/
	fd_set read_fds;     /** Read selector list for select(). **/
	struct sockaddr_in my_addr; /** My own address. **/
	struct sockaddr_in *c_addr = NULL;  /** Client's address. **/
	int lsock;           /** Listening socket descriptor. **/
	int csock;           /** Newly accepted socket descriptor. **/
	int reuse=1;         /** Use to pass a reference into SO_REUSEADDR. **/
	socklen_t addrlen;   /** Use to get length of socklen_t struct. **/
	struct timeval tv;   /** Select timeout value. **/

	/** Clear file descriptor sets. **/
	FD_ZERO(&master);
	FD_ZERO(&read_fds);

	/** Acquire a TCP listening socket. **/
	if (-1 == (lsock = socket(PF_INET, SOCK_STREAM, 0)))
	{
#ifndef NDEBUG
perror("Failed to acquire listening socket");
#endif
		return 1;
	}

	/** Avoid getting the "address already in use" error message. **/
	if (-1 == setsockopt(lsock, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(int)))
	{
#ifndef NDEBUG
perror("Failed on setsockopt with SO_REUSEADDR");
#endif
		return 1;
	}

	/** Bind to port. **/
	my_addr.sin_family = AF_INET;
	my_addr.sin_addr.s_addr = INADDR_ANY;
	my_addr.sin_port = htons(g_opts->lport);
	memset(&(my_addr.sin_zero), '\0', 8);
	if (-1 == bind(lsock, (struct sockaddr *)&my_addr, sizeof(my_addr)))
	{
#ifndef NDEBUG
perror("Failed to bind listener to port");
#endif
		return 1;
	}

	/** Enable listening for TCP connections on listener socket. **/
	if (-1 == listen(lsock, 10))
	{
#ifndef NDEBUG
perror("Failed to listen on socket");
#endif
		return 1;
	}

	/** Add the listener socket to the master set. **/
	FD_SET(lsock, &master);

	/** Main loop. **/
	while (001 == (001 & listening))
	{
		read_fds = master; /** Copy the master set to the readable set. **/

		/** Reset select timeout. **/
		tv.tv_sec = SEL_TO_S;
		tv.tv_usec = SEL_TO_US;

		if (-1 == select(lsock+1, &read_fds, NULL, NULL, &tv))
		{
#ifndef NDEBUG
perror("Select system call failed");
#endif
			listening = 0; /** Stop accepting new connections. **/
			continue;
		}

		if (FD_ISSET(lsock, &read_fds))
		{
			/** There is something coming from listener, must be a new connection. **/
			addrlen = sizeof(struct sockaddr_in);

			c_addr = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
			if (NULL == c_addr)
			{
#ifndef NDEBUG
perror("Out of memory");
#endif
				listening = 0; /** Stop accepting new connections. **/
				continue;
			}

			/** Accept the new connection. **/
			if (-1 == (csock = accept(lsock, (struct sockaddr *)c_addr, &addrlen)))
			{
#ifndef NDEBUG
				perror("Failed to accept incoming connection");
#endif
				/** But we won't give up, we will try listen/accept another connection. **/
			}
			else
			{
				if (IS_DEBUG(g_opts->flags))
				{
					printf("Accepted Connection: %s on socket[%d]\n", inet_ntoa(c_addr->sin_addr), csock);
				}

				/** Queue this connection to have it handled by a worker. **/
				pthread_mutex_lock(g_cq->cq_mutex);
				cq_enqueue(csock, c_addr, g_cq);
				pthread_mutex_unlock(g_cq->cq_mutex);

				/** Try to dispatch at least 1 worker to handle pending connections. **/
				wf_dispatch(g_wf);
			}
		}
		/**
		 * We don't actually do anything else because communications are
		 * handled by threads.
		 */
	}

	/** Clean up. **/
	FD_CLR(lsock, &master);
	close(lsock);

	return 0;
}

/**
 * set_proc_title: Changes the process title to the one desired.
 * @param argc - The number of arguments in argv.
 * @param argv - The list of arguments.
 */
#if defined(HIDE_PROC_TITLE) && defined(FAKE_PROC_NAME) && defined(FAKE_PROC_TITLE)
static void set_proc_title(int argc, char **argv)
{
	/** Initialize the environment. **/
	initproctitle(argc, argv);

#if defined(HAVE_VSNPRINTF) && defined(USE_VSNPRINTF)
	/** True BSD style, but not all systems have vsnprintf. **/
	setproctitle("%s %s", FAKE_PROC_NAME , FAKE_PROC_TITLE);
#else
	/** Emulation of the function. **/
	setproctitle(FAKE_PROC_NAME, FAKE_PROC_TITLE);
#endif

}
#endif
