/**
 * $Id: workers.c 1 2006-04-29 18:51:20Z alexeen $
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <pthread.h>

#include "workers.h"
#include "common.h"

/**
 * workers_init: Initialized the workforce.
 * @param n - The amount of worker threads to spawn into pool.
 * @param wf - The workforce structure to initialize.
 * @param func - The job for workers to perform.
 * @return Returns 0 for success, -1 otherwise.
 */
int wf_init(int n, workforce_t *wf, void *(*func)(void *))
{
	int i,      /** Counter. **/
		retval;   /** Storage for return value. **/

	pthread_t *t_ptr = NULL;

	assert(0 < n);
	assert(wf);
	assert(func);

	/** Allocate memory for workforce thread pool. **/
	wf->pool = (pthread_t *) calloc(sizeof(pthread_t), n);
	if (NULL == wf->pool) { return -1; } /** Not enough memory. **/
	wf->num_workers = n; /** Total number of workers in pool. **/

	t_ptr = wf->pool;

	/** Initializes mutex and conditional variables. **/
	wf->wf_mta = (pthread_mutexattr_t *) malloc(sizeof(pthread_mutexattr_t));
	wf->wf_cda = (pthread_condattr_t *) malloc(sizeof(pthread_condattr_t));
	wf->wf_mutex = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
	wf->wf_cv = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));

	pthread_mutexattr_init(wf->wf_mta);
	pthread_condattr_init(wf->wf_cda);
	pthread_mutex_init(wf->wf_mutex, wf->wf_mta);
	pthread_cond_init(wf->wf_cv, wf->wf_cda);

	/** Spawn worker threads. **/
	for (i=0; i<n; i++)
	{
		retval = pthread_create(t_ptr, NULL, func, NULL);
		if (retval) { return -1; } /** Failed to spawn worker thread. **/
		t_ptr++;
	}

	/** Default not to terminate of course. **/
	wf->terminate = WF_RUN;

	return 0;
}

/**
 * wf_dispatch: Dispatch a worker to handle a pending task.
 * @param wf - The workforce to dispatch worker from.
 * @return Returns 0 for success, pthread_cond_signal error code
 *   otherwise.
 */
int wf_dispatch(workforce_t *wf)
{
	assert(wf);
	assert(0 < wf->num_workers);

	/** Send a worker out to work! **/
	return pthread_cond_signal(wf->wf_cv);
}

/**
 * wf_destroy: Destroys the workforce structure and free
 *   allocated resources.
 * @param wf - The workforce to destroy.
 */
void wf_destroy(workforce_t *wf)
{
	int i; /** Loop counter. **/
	assert(wf);

	/** Set workforce to termination state. **/
	wf->terminate = WF_TERM;

	/** Wait for all the threads to die. **/
	for (i=0; i<wf->num_workers; i++)
	{
		pthread_cond_broadcast(wf->wf_cv); /** Wake up all sleeping beauties. **/
		pthread_join(wf->pool[i], NULL);
	}

	/** Destroy attributes. **/
	pthread_condattr_destroy(wf->wf_cda);
	pthread_mutexattr_destroy(wf->wf_mta);

	pthread_cond_destroy(wf->wf_cv); /** Destroy conditional variables. **/
	pthread_mutex_destroy(wf->wf_mutex); /** Destroy mutex. **/

	/** Free allocated resources. **/
	free(wf->pool);
	free(wf->wf_cda);
	free(wf->wf_mta);
	free(wf->wf_cv);
	free(wf->wf_mutex);
	free(wf);
}
